import os
import json
import time
from collections import namedtuple
from threading import Thread, Event
from .datastore import DeepspeedDatastore, _best_effort_read_key

TaskStatus = namedtuple("TaskStatus", ["status", "timestamp", "node_index"])


class TASK_STATUS:
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"
    UNREACHABLE = "unreachable"


class TaskStatusNotifier:
    ROOT = "node_status"

    def __init__(
        self,
        datastore: DeepspeedDatastore,
    ):
        self._datastore = datastore

    def _path(self, node_index):
        return os.path.join(self.ROOT, str(node_index))

    def _create_status(self, status, node_index):
        return json.dumps(
            {"status": status, "timestamp": time.time(), "node_index": node_index}
        )

    def _notify(self, status: str, node_index: int):
        self._datastore.put(
            self._path(node_index),
            self._create_status(status, node_index),
            overwrite=True,
        )

    def heartbeat(self, node_index: int):
        self._notify(TASK_STATUS.RUNNING, node_index)

    def running(self, node_index: int):
        self._notify(TASK_STATUS.RUNNING, node_index)

    def finished(self, node_index: int):
        self._notify(TASK_STATUS.FINISHED, node_index)

    def failed(self, node_index: int):
        self._notify(TASK_STATUS.FAILED, node_index)

    def read(self, node_index: int) -> TaskStatus:
        data = _best_effort_read_key(self._datastore, self._path(node_index))
        if not data:
            return TaskStatus(
                status=TASK_STATUS.UNREACHABLE, timestamp=None, node_index=node_index
            )
        data = json.loads(data.text)
        return TaskStatus(
            status=data["status"],
            timestamp=data["timestamp"],
            node_index=data["node_index"],
        )


class TaskUnreachableException(Exception):
    pass


class TaskFailedException(Exception):
    pass


class HeartbeatTimeoutException(Exception):
    pass


def wait_for_task_to_be_reachable(
    task_status_notifier: TaskStatusNotifier, node_index: int, timeout: int
):
    while task_status_notifier.read(node_index).status == TASK_STATUS.UNREACHABLE:
        time.sleep(1)
        timeout -= 1
        if timeout <= 0:
            # TODO: Improve timeout error messages to make it clearer to the user that
            # the task was unreachable and we hit a timeout for that.
            raise TimeoutError


def wait_for_task_completion(
    task_status_notifier: TaskStatusNotifier,
    node_index: int,
    heartbeat_timeout: int = 60 * 60,
    unreachable_timeout: int = 60 * 5,
):
    wait_for_task_to_be_reachable(task_status_notifier, node_index, unreachable_timeout)
    while True:
        status = task_status_notifier.read(node_index)
        if status.status == TASK_STATUS.FINISHED:
            return
        if status.status == TASK_STATUS.FAILED:
            raise TaskFailedException
        if time.time() - status.timestamp > heartbeat_timeout:
            raise HeartbeatTimeoutException
        time.sleep(1)


class HeartbeatThread(Thread):
    def __init__(
        self,
        task_status_notifier: TaskStatusNotifier,
        node_index: int,
        heartbeat_interval: int = 10,
    ):
        super().__init__()
        self.task_status_notifier = task_status_notifier
        self.node_index = node_index
        self.heartbeat_interval = heartbeat_interval
        self._exit_event = Event()

    def run(self):
        while self._exit_event.is_set() is False:
            self.task_status_notifier.heartbeat(self.node_index)
            time.sleep(self.heartbeat_interval)

    def stop(self):
        self._exit_event.set()
        self.join()



HOSTFILE_IP_KEY = "hostfile_ips"
HOSTFILE = "hostfile.txt"
DEEPSPEED_JOB_COMPLETE_VAR = "mpi_job_status"
PUBLIC_KEY_RECEIVED_VAR = "public_key_received"
CONTROL_TASK_DONE_PATH = "control"
MPI_PUBLIC_KEY_PATH = "public_keys"
DEEPSPEED_ENV_FILE = ".deepspeed_env"  # https://github.com/microsoft/DeepSpeed/blob/24f20ef0a105d32f6085fe0d3b1c2f9324a6262c/docs/_tutorials/getting-started.md?plain=1#L230-L254
DEEPSPEED_SUFFIX = "mf.deepspeed"


__mf_extensions__ = "deepspeed"



try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution("metaflow-deepspeed").version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None



import datetime
import json
from functools import wraps
from collections import defaultdict
import tempfile
import os
from .constants import DEFAULT_FILE_NAME, DEFAULT_PROFILER_FILE_NAME

try:
    from transformers import TrainerCallback
except ImportError:

    class TrainerCallback:
        pass

class MetaflowHuggingFaceCardCallback(TrainerCallback):
    """
    A basic Huggingface trainer callback that updates a Metaflow card components.
    It updates progress bars, sets parameters, logs metrics, and final outputs.
    """

    MODEL_CARD_ID = "huggingface_model_card"

    DEFAULT_FILE_NAME = DEFAULT_FILE_NAME

    _metrics = defaultdict(list)

    def _save(self):
        save_data = {
            "metrics": self._metrics,
            "model_config": self._model_config,
            "trainer_configuration": self._trainer_configuration,
            "runtime_info": self._runtime_info,
            "training_state": self._state,
            "created_on": datetime.datetime.now().isoformat(),
        }
        with open(self._save_file_name, "w") as f:
            json.dump(save_data, f)

    def _resolve_save_file_name(self, save_file_name):
        if save_file_name is not None:
            return save_file_name
        if os.environ.get("METAFLOW_HF_CARD_SAVE_FILE_NAME", None):
            return os.environ["METAFLOW_HF_CARD_SAVE_FILE_NAME"]
        save_file_name = self.DEFAULT_FILE_NAME
        return save_file_name

    def __init__(
        self, 
        tracked_metrics=["loss", "grad_norm"], 
        save_file_name=None,
    ) -> None:
        super().__init__()
        
        self.tracked_metrics = tracked_metrics
        self._save_file_name = self._resolve_save_file_name(save_file_name)

    def on_train_begin(
        self,
        args,
        state,
        control,
        model=None,
        **kwargs,
    ):
        self._model_config = None if model is None else model.config.to_dict()
        self._trainer_configuration = args.to_dict()
        self._runtime_info = {}
        self._state = {
            "epoch": state.epoch,
            "global_step": state.global_step,
            "max_steps": state.max_steps,
            "num_train_epochs": state.num_train_epochs,
        }
        self._training_state = {}

    def on_log(self, args, state, control, model=None, logs=None, **kwargs):
        if not state.is_local_process_zero:
            return
        self._state = {
            "epoch": state.epoch,
            "global_step": state.global_step,
            "max_steps": state.max_steps,
            "num_train_epochs": state.num_train_epochs,
        }
        for metric in logs:
            if metric in self.tracked_metrics:
                data = {"step": state.global_step, "value": logs[metric]}
                self._metrics[metric].append(data)

        if "train_runtime" in logs:
            self._runtime_info = {
                "Train runtime": logs["train_runtime"],
                "Train samples / sec": logs["train_samples_per_second"],
                "Train steps / sec": logs["train_steps_per_second"],
                # "Total Floating point operations": logs["total_flos"],
                # "TFLOPs": logs["total_flos"] // 1e12 / logs["train_runtime"],
            }
        self._save()


import os
from io import BytesIO
from itertools import starmap
from collections import namedtuple
from typing import List, Optional, Union, Tuple

from metaflow import FlowSpec
from metaflow.metaflow_config import DATATOOLS_AZUREROOT, DATASTORE_SYSROOT_AZURE
from metaflow.plugins.azure.blob_service_client_factory import (
    get_azure_blob_service_client,
)
from metaflow.plugins.datastores.azure_storage import AzureStorage, _AzureRootClient
from metaflow.exception import MetaflowException


AzureObject = namedtuple("AzureObject", "blob url key text")
AzureListPathResult = namedtuple("AzureListPathResult", "url key")


class AzureDatastoreNotFoundError(MetaflowException):
    headline = "Azure Datastore Object Not Found"

    def __init__(self, url):
        super(AzureDatastoreNotFoundError, self).__init__(
            "The Azure Datastore object at '%s' was not found." % url
        )


class AzureBlob(object):

    """
    An experimental class to interact with Azure Blob storage.
    It aims to mimic a slice of the Metaflow S3 client API, including these methods:
        - put: place a string or bytes object into a key in the datastore.
        - get: retrieve bytes or a string from a key in the datastore.
        - list_paths: list all objects under the datastore's `keys`.
        - put_files: modified to put raw bytes instead of files.
        - get_files (new): use to download a list of files from the datastore that were put there using put_files.
        - get_recursive: get all objects under the datastore's `keys`.
    """

    def __init__(
        self,
        run: Optional[Union[FlowSpec, "metaflow.Run"]] = None,
        blob_path: Optional[str] = None,
        run_pathspec: Optional[str] = None,
        step_name: Optional[str] = None,
    ):
        if blob_path is None and run_pathspec is None:
            raise ValueError(
                "Either blob_path or run_pathspec must be specified in the AzureBlob instantiation."
            )
        if blob_path and run_pathspec:
            raise ValueError(
                "Cannot specify both blob_path and run_pathspec in the AzureBlob instantiation."
            )

        # TODO: tmp
        if blob_path:
            raise NotImplementedError(
                "Custom blob_path is not yet implemented in AzureBlob. Pass flow_name and run_id instead."
            )

        self._datastore = AzureStorage()
        self._datastore_root = DATASTORE_SYSROOT_AZURE
        self._datastore.datastore_root = self._datastore_root
        self._run_pathspec = run_pathspec
        self._blob_path = blob_path
        self._step_name = step_name

    def get_datastore_key_location(self, key: Optional[str] = None):
        if self._step_name is None:
            return os.path.join(
                self._datastore_root, self._run_pathspec, key if key else ""
            ).rstrip("/")
        else:
            return os.path.join(
                self._datastore_root, self._run_pathspec, self._step_name, key if key else ""
            ).rstrip("/")

    def put(self, key: str, obj: Union[str, bytes], overwrite: bool = False):
        "Put a single object into the datastore's `key` index."

        if isinstance(obj, bytes):
            self._datastore.save_bytes(
                [(self.get_datastore_key_location(key), BytesIO(obj))],
                overwrite=overwrite,
            )
        else:
            self._datastore.save_bytes(
                [(self.get_datastore_key_location(key), BytesIO(obj.encode("utf-8")))],
                overwrite=overwrite,
            )

    def get(self, key):
        "Get a single object residing in the datastore's `key` index."
        datastore_url = self.get_datastore_key_location(key)
        keyless_root = self.get_datastore_key_location()
        with self._datastore.load_bytes([datastore_url]) as get_results:
            for key, path, meta in get_results:
                if path is not None:
                    with open(path, "rb") as f:
                        blob_bytes = f.read()
                        try:
                            return AzureObject(
                                blob=blob_bytes,
                                url=datastore_url,
                                text=blob_bytes.decode("utf-8"),
                                key=datastore_url[len(keyless_root) :].strip("/"),
                            )
                        except UnicodeDecodeError:
                            return AzureObject(
                                blob=blob_bytes,
                                url=datastore_url,
                                text=None,
                                key=datastore_url[len(keyless_root) :].strip("/"),
                            )
                else:
                    raise AzureDatastoreNotFoundError(datastore_url)

    def list_paths(self, keys: Optional[List[str]] = None):
        "List all objects in the datastore's `keys` index."
        keyless_root = self.get_datastore_key_location()
        if keys is None:
            keys = [keyless_root]
        else:
            keys = [self.get_datastore_key_location(key) for key in keys]
        results = []
        for list_content_result in self._datastore.list_content(keys):
            key_suffix = list_content_result.path[len(keyless_root) :].strip("/")
            if list_content_result.is_file:
                results.append(
                    AzureListPathResult(url=list_content_result.path, key=key_suffix)
                )
            else:
                results += self.list_paths([key_suffix])
        return results

    def put_files(self, key_paths: List[Tuple[str, str]], overwrite=False):
        results = []
        for key, path in key_paths:
            with open(path, "rb") as f:
                self.put(key, f.read(), overwrite=overwrite)
            results.append(self.get_datastore_key_location(key))
        return results

    def get_files(self, key_paths: List[Tuple[str, str]], as_binary=False):
        for key, path in key_paths:
            if not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))

            if as_binary:
                with open(path, "wb") as f:
                    f.write(self.get(key).blob)
            else:
                with open(path, "w") as f:
                    f.write(self.get(key).text)



from metaflow.exception import MetaflowException


class DeepspeedException(MetaflowException):
    headline = "Deepspeed Command Failed"

    def __init__(self, msg):
        super(DeepspeedException, self).__init__(msg)


class DatastoreKeyNotFoundError(MetaflowException):
    headline = "DeepSpeed Datastore Not Found"

    def __init__(self, datastore_path_name):
        msg = "The DeepSpeed datastore path {} was not found.".format(
            datastore_path_name
        )
        super(DatastoreKeyNotFoundError, self).__init__(msg)


class BarrierTimeoutException(MetaflowException):
    headline = "Barrier Timeout"

    def __init__(self, lock_name, description):
        msg = f"Task has timed out after waiting for some keys to be written to the datastore.\n[Barrier Name]:{lock_name}\n[Barrier Info]: {description}"
        super(BarrierTimeoutException, self).__init__(msg)


class SSHKeyGenException(MetaflowException):
    headline = "SSH Key Generation Failed"

    def __init__(self, cmd, exception, worker_type, node_index):
        msg = f"Failed to generate SSH keys for the worker type {worker_type} and node index {node_index}. \n\n[Command]: {cmd}\n\n[Exception]: {exception}"
        super(SSHKeyGenException, self).__init__(msg)


class SSHServiceRestartException(MetaflowException):
    headline = "SSH Service Restart Failed"

    def __init__(self, cmd, exception, worker_type, node_index):
        msg = f"Failed to restart the SSH service for the worker type {worker_type} and node index {node_index}. \n\n[Command]: {cmd}\n\n[Exception]: {exception}"
        super(SSHServiceRestartException, self).__init__(msg)


class SSHScanUnsuccessfulException(MetaflowException):
    headline = "`ssh-keyscan` Unsuccessful"

    def __init__(
        self,
        bad_host,
        keyscan_error,
        ubf_context,
        node_index,
    ):
        _err = "[couldn't capture standard-error]"
        if keyscan_error is not None:
            _err = keyscan_error
        msg = f"Failed to capture the SSH keys using `ssh-keyscan` for the worker type {ubf_context} and node index {node_index}. `ssh-keyscan` failures indicate that the nodes are unable to communicate with each other via ssh. \n\n[Error]: {_err}\n\n[Bad Host]: {bad_host}"
        super(SSHScanUnsuccessfulException, self).__init__(msg)



from metaflow.metaflow_current import current
import json
from threading import Thread, Event
import time


class InfoCollectorThread(Thread):
    def __init__(
        self,
        interval=1,
        file_name=None,
    ):
        super().__init__()
        self._exit_event = Event()
        self._interval = interval
        assert file_name is not None, "file_name must be provided"
        self._file_name = file_name
        self.daemon = True
        self._data = {}
        self._has_errored = False
        self._current_error = None

    def read(self):
        return self._data

    def has_errored(self):
        return self._has_errored

    def get_error(self):
        return self._current_error

    def _safely_load(self):
        try:
            with open(self._file_name, "r") as f:
                return json.load(f), None
        except FileNotFoundError as e:
            return {}, e
        except Exception as e:
            return {}, e

    def run(self):
        while self._exit_event.is_set() is False:
            data, self._current_error = self._safely_load()
            if not self._current_error:
                self._data = data
            self._has_errored = True if self._current_error else False
            time.sleep(self._interval)

    def stop(self):
        self._exit_event.set()
        self.join()


class CardRefresher:

    CARD_ID = None

    def on_startup(self, current_card):
        raise NotImplementedError("make_card method must be implemented")

    def on_error(self, current_card, error_message):
        raise NotImplementedError("error_card method must be implemented")

    def on_update(self, current_card, data_object):
        raise NotImplementedError("update_card method must be implemented")


class CardUpdaterThread(Thread):
    def __init__(
        self,
        card_refresher: CardRefresher,
        interval=1,
        file_name=None,
        collector_thread: InfoCollectorThread = None,
    ):
        super().__init__()
        self._exit_event = Event()
        self._interval = interval
        self._refresher = card_refresher
        self._file_name = file_name
        self._collector_thread = collector_thread
        self.daemon = True

    def run(self):
        if self._refresher.CARD_ID is None:
            raise ValueError("CARD_ID must be defined")
        current_card = current.card[self._refresher.CARD_ID]
        self._refresher.on_startup(current_card)
        while self._exit_event.is_set() is False:
            data = self._collector_thread.read()
            if self._collector_thread.has_errored():
                self._refresher.on_error(
                    current_card, self._collector_thread.get_error()
                )
            self._refresher.on_update(current_card, data)
            time.sleep(self._interval)

    def stop(self):
        self._exit_event.set()
        self._collector_thread.stop()
        self.join()



class AsyncPeriodicRefresher:

    def __init__(
        self,
        card_referesher: CardRefresher,
        updater_interval=1,
        collector_interval=1,
        file_name=None,
    ):
        assert card_referesher.CARD_ID is not None, "CARD_ID must be defined"
        self._collector_thread = InfoCollectorThread(
            interval=collector_interval, file_name=file_name
        )
        self._collector_thread.start()
        self._updater_thread = CardUpdaterThread(
            card_refresher=card_referesher,
            interval=updater_interval,
            file_name=file_name,
            collector_thread=self._collector_thread,
        )

    def start(self):
        self._updater_thread.start()

    def stop(self):
        self._updater_thread.stop()
        self._collector_thread.stop()



from typing import Dict, Union
import os
import tarfile
from metaflow import Run, S3
# from metaflow_extensions.deepspeed.plugins.azure_blob import AzureBlob
from metaflow.plugins.az_store import AzureBlob

from transformers import TrainerCallback


class DeepspeedHFTrainerS3Sync(TrainerCallback):

    def __init__(
        self,
        run_pathspec=None,
        s3_root=None,
        training_outputs_path="training_outputs",
        local_rank=None,
        node_index=None,
        push_from_all_nodes=False,  
    ):
        self.root = training_outputs_path
        self.run = Run(run_pathspec)
        self.s3_root = s3_root
        self.local_rank = local_rank
        self.node_index = node_index
        self.push_from_all_nodes = push_from_all_nodes

    def on_epoch_end(self, args, state, control, **kwargs):
        """
        Push the training outputs to Metaflow S3 on epoch end.
        Find other hooks at https://huggingface.co/docs/transformers/v4.38.1/en/main_classes/callback#transformers.TrainerCallback

        NOTE: As models get larger, be careful the amount of checkpoints you push to S3. You may, for example, only want to push the best checkpoint.
        """
        if os.path.exists(self.root) and (
            self.local_rank is not None and self.local_rank == 0
        ):
            try:
                if self.push_from_all_nodes or self.node_index == 0:
                    print(f"Pushing {self.root} to S3 on node {self.node_index}.")
                    tar_bytes = self._get_tar_bytes()
                    self._upload_to_s3(tar_bytes)
                    print(f"Pushed to {self.s3_path} on node {self.node_index}.")
            except Exception as e:
                print(f"Error pushing to S3 on node {self.node_index}: {e}")

    def _get_tar_bytes(self):
        "Zip from the root of the directory and return the bytes of the tar file."
        with tarfile.open(f"{self.root}.tar.gz", "w:gz") as tar:
            tar.add(self.root, arcname=os.path.basename(self.root))
        with open(f"{self.root}.tar.gz", "rb") as f:
            tar_bytes = f.read()
        return tar_bytes

    def _get_s3_client(self):
        "Return an S3 object based on the run or s3_root."
        if self.run:
            return S3(run=self.run)
        elif self.s3_root:
            return S3(s3root=self.s3_root)
        else:
            return S3(s3root=os.path.join(DATATOOLS_S3ROOT, self.root))

    def _upload_to_s3(self, tar_bytes):
        "Push the tar file to S3."
        s3 = self._get_s3_client()
        if s3 is None:
            return None
        if self.local_rank is not None and self.local_rank == 0:
            self.s3_path = s3.put(
                f"{self.root}-node-{self.node_index}.tar.gz", tar_bytes
            )
        else:
            self.s3_path = s3.put(f"{self.root}.tar.gz", tar_bytes)
        s3.close()

    def _download_from_s3(
        self, all_nodes: bool = False
    ) -> Union[bytes, Dict[str, bytes]]:
        "Pull the tar file(s) from S3."
        s3 = self._get_s3_client()
        candidate_paths = s3.list_paths()
        if all_nodes:
            tar_balls = {}
            for s3obj in candidate_paths:
                if self.root in s3obj.key:
                    obj = s3.get(s3obj.key)
                    tar_balls[obj.key] = obj.blob
            s3.close()
            return tar_balls
        elif self.node_index is not None:
            tar_bytes = s3.get(f"{self.root}-node-{self.node_index}.tar.gz").blob
        else:
            tar_bytes = s3.get(f"{self.root}.tar.gz").blob
        s3.close()
        return tar_bytes

    def _extract_tar(self, tar_bytes, path=None):
        """
        Extract the tar file to the root of the directory.
        If `path` is specified, assumed to be a file path and extract to that location.
        The use case for path is when downloading all checkpoints from many nodes nodes.
        """
        if path:
            with open(path, "wb") as f:
                f.write(tar_bytes)
            with tarfile.open(path, "r:gz") as tar:
                tar.extractall(path=path.replace(".tar.gz", ""))
            os.remove(path)
        else:
            with open(f"{self.root}.tar.gz", "wb") as f:
                f.write(tar_bytes)
            with tarfile.open(f"{self.root}.tar.gz", "r:gz") as tar:
                tar.extractall(path=os.path.dirname(self.root))
            os.remove(f"{self.root}.tar.gz")

    def download(self, all_nodes=False):
        if all_nodes:
            tar_balls = self._download_from_s3(all_nodes=all_nodes)
            for _path, _bytes in tar_balls.items():
                self._extract_tar(_bytes, path=_path)
        else:
            tar_bytes = self._download_from_s3()
            self._extract_tar(tar_bytes)


class DeepspeedHFTrainerAzureBlobSync(TrainerCallback):

    def __init__(
        self,
        run_pathspec=None,
        training_outputs_path="training_outputs",
        local_rank=None,
        node_index=None,
        push_from_all_nodes=False
    ):
        self.root = training_outputs_path
        self.run_pathspec = run_pathspec
        self.local_rank = local_rank
        self.node_index = node_index
        self.push_from_all_nodes = push_from_all_nodes

    def on_epoch_end(self, args, state, control, **kwargs):
        """
        Push the training outputs to Metaflow Azure Blob Storage on epoch end.
        Find other hooks at https://huggingface.co/docs/transformers/v4.38.1/en/main_classes/callback#transformers.TrainerCallback

        NOTE: As models get larger, be careful the amount of checkpoints you push to Azure Blob. You may, for example, only want to push the best checkpoint.
        """
        if os.path.exists(self.root) and (
            self.local_rank is not None and self.local_rank == 0
        ):
            try:
                if self.push_from_all_nodes or self.node_index == 0:
                    print(f'Zipping and pushing {self.root} to Azure Blob.')
                    tar_bytes = self._get_tar_bytes()
                    self._upload_to_azure_blob(tar_bytes)
                    print(f"Pushed to Azure Blob.")
            except Exception as e:
                print(f"Error pushing to Azure Blob: {e}")

    def _get_tar_bytes(self):
        "Zip from the root of the directory and return the bytes of the tar file."
        with tarfile.open(f"{self.root}.tar.gz", "w:gz") as tar:
            tar.add(self.root, arcname=os.path.basename(self.root))
        with open(f"{self.root}.tar.gz", "rb") as f:
            tar_bytes = f.read()
        return tar_bytes

    def _get_blob_store(self):
        "Return an Azure Blob object based on the run."
        return AzureBlob(run_pathspec=self.run_pathspec)

    def _upload_to_azure_blob(self, tar_bytes):
        "Push the tar file to Azure Blob."
        blob_store = self._get_blob_store()
        if self.local_rank is not None and self.local_rank == 0:
            blob_store.put(
                f"{self.root}-node-{self.node_index}.tar.gz", tar_bytes
            )
        else:
            blob_store.put(f"{self.root}.tar.gz", tar_bytes)

    def _download_from_azure_blob(
        self, all_nodes: bool = False
    ) -> Union[bytes, Dict[str, bytes]]:
        "Pull the tar file(s) from Azure Blob."
        blob_store = self._get_blob_store()
        candidate_paths = blob_store.list_paths()
        if all_nodes:
            tar_balls = {}
            for obj in candidate_paths:
                if self.root in obj.key:
                    obj = blob_store.get(obj.key)
                    tar_balls[obj.key] = obj.blob
            return tar_balls
        elif self.node_index is not None:
            tar_bytes = blob_store.get(f"{self.root}-node-{self.node_index}.tar.gz").blob
        else:
            tar_bytes = blob_store.get(f"{self.root}.tar.gz").blob_bytes
        return tar_bytes

    def _extract_tar(self, tar_bytes, path=None):
        """
        Extract the tar file to the root of the directory.
        If `path` is specified, assumed to be a file path and extract to that location.
        The use case for path is when downloading all checkpoints from many nodes nodes.
        """
        if path:
            with open(path, "wb") as f:
                f.write(tar_bytes)
            with tarfile.open(path, "r:gz") as tar:
                tar.extractall(path=path.replace(".tar.gz", ""))
            os.remove(path)
        else:
            with open(f"{self.root}.tar.gz", "wb") as f:
                f.write(tar_bytes)
            with tarfile.open(f"{self.root}.tar.gz", "r:gz") as tar:
                tar.extractall(path=os.path.dirname(self.root))
            os.remove(f"{self.root}.tar.gz")

    def download(self, all_nodes=False):
        if all_nodes:
            tar_balls = self._download_from_azure_blob(all_nodes=all_nodes)
            for _path, _bytes in tar_balls.items():
                self._extract_tar(_bytes, path=_path)
        else:
            tar_bytes = self._download_from_azure_blob()
            self._extract_tar(tar_bytes)


STEP_DECORATORS_DESC = [
    ("deepspeed", ".deepspeed_libs.deepspeed_decorator.DeepspeedDecorator"),
    ("huggingface_card", ".deepspeed_libs.hugging_face.card_decorator.HuggingFaceCardDecorator"),
]
__mf_promote_submodules__ = ["az_store", "hf_callbacks", "deepspeed_libs",]


from typing import List, Dict, Union, Tuple
import time
import sys
import os
from io import BytesIO
from collections import namedtuple
from contextlib import contextmanager
from .exceptions import DatastoreKeyNotFoundError, BarrierTimeoutException
from .constants import DEEPSPEED_SUFFIX

# mimic a subset of the behavior of the Metaflow S3Object
DeepspeedDatastoreBlob = namedtuple("DeepspeedDatastoreBlob", "blob url text")
DeepspeedListPathResult = namedtuple("DeepspeedListPathResult", "url")


class DeepspeedDatastore(object):
    """
    This class is a wrapper around the basic Metaflow cloud datastore functionality.
    It is used to interact with each cloud datastore provider from within the DeepspeedExecutor and DeepspeedDecorator class.
    For now, local storage is not supported.
    Methods provided follow the naming convention of Metaflow's S3 client: put, get, and list_paths.
    """

    def __init__(self, flow_datastore, pathspec=None):
        self._backend = flow_datastore._storage_impl
        self._flow_name = flow_datastore.flow_name
        _, run_id, step_name, _ = pathspec.split("/")
        self._run_id = run_id
        self._step_name = step_name

    @property
    def get_storage_root(self):
        """
        Return the path to the root of the deepspeed datastore.
        This method is where the unique deepspeed datastore root for each cloud provider is specified.

        Note: S3Storage class uses the S3 client (other clouds do not have this),
            which prepends the storage root inside the self._backend calls this class uses.
        """
        if self._backend.TYPE == "s3":
            return DEEPSPEED_SUFFIX
        elif self._backend.TYPE == "azure":
            from metaflow.metaflow_config import DATASTORE_SYSROOT_AZURE

            return os.path.join(DATASTORE_SYSROOT_AZURE, DEEPSPEED_SUFFIX)
        elif self._backend.TYPE == "gs":
            from metaflow.metaflow_config import DATASTORE_SYSROOT_GS

            return os.path.join(DATASTORE_SYSROOT_GS, DEEPSPEED_SUFFIX)
        else:
            raise NotImplementedError(
                "Deepspeed datastore does not support backend %s" % (self._backend.TYPE)
            )

    def get_datastore_key_location(self):
        return os.path.join(
            self.get_storage_root, self._flow_name, self._run_id, self._step_name
        )

    def get_datastore_file_location(self, key):
        return os.path.join(self.get_datastore_key_location(), key)

    def put(self, key: str, obj: Union[str, bytes], overwrite: bool = False):
        "Put a single object into the datastore's `key` index."
        _save_object = None
        if isinstance(obj, bytes):
            _save_object = BytesIO(obj)
        else:
            _save_object = BytesIO(obj.encode("utf-8"))

        self._backend.save_bytes(
            [(self.get_datastore_file_location(key), _save_object)],
            overwrite=overwrite,
        )

    def put_files(self, key_paths: List[Tuple[str, str]], overwrite=False):
        results = []
        for key, path in key_paths:
            with open(path, "rb") as f:
                self.put(key, f.read(), overwrite=overwrite)
            results.append(self.get_datastore_file_location(key))
        return results

    def get(self, key):
        "Get a single object residing in the datastore's `key` index."
        datastore_url = self.get_datastore_file_location(key)
        with self._backend.load_bytes([datastore_url]) as get_results:
            for key, path, _ in get_results:
                if path is not None:
                    with open(path, "rb") as f:
                        blob_bytes = f.read()
                        return DeepspeedDatastoreBlob(
                            blob=blob_bytes,
                            url=datastore_url,
                            text=blob_bytes.decode("utf-8"),
                        )
                else:
                    raise DatastoreKeyNotFoundError(datastore_url)

    def get_many(self, keys):
        return [self.get(key) for key in keys]

    def list_paths(self, keys):
        "List all objects in the datastore's `keys` index."
        if self._backend.TYPE == "gs":
            raise NotImplementedError(
                """Deepspeed datastore does not support the list_paths API for Google Cloud storage. 
                If you know the paths ahead of this call, use get_many with the keys you want to list instead."""
            )
        keys = [self.get_datastore_file_location(key) for key in keys]
        list_path_results = [
            DeepspeedListPathResult(url=list_content_result.path)
            for list_content_result in self._backend.list_content(keys)
        ]
        return list_path_results


def _best_effort_read_key(datastore: DeepspeedDatastore, key: str):
    """
    Read a key from the datastore, but do not raise an error if the key is not found.
    """
    try:
        return datastore.get(key)
    except DatastoreKeyNotFoundError:
        return None


def _best_effort_get_keys(datastore: DeepspeedDatastore, keys: List[str]):
    """
    Get the keys from the datastore, but do not raise an error if the keys are not found.
    """
    results = {}
    not_found_keys = []
    for key in keys:
        data = _best_effort_read_key(datastore, key)
        if data:
            results[key] = data
        else:
            not_found_keys.append(key)
    return results, not_found_keys


def wait_for_key_data(
    datastore: DeepspeedDatastore,
    keys: List[str],
    max_wait_time: float = 600,
    frequency=0.1,
) -> Dict[str, DeepspeedDatastoreBlob]:
    """
    Wait for the keys to be available in the datastore.
    If the keys are not available after `max_wait_time` seconds, raise an error.
    """
    start = time.time()
    exit_condition = lambda: time.time() - start > max_wait_time
    _current_keys = keys.copy()
    main_data = {}
    while not exit_condition():
        data, _ = _best_effort_get_keys(datastore, _current_keys)
        # if all keys are found, return the data
        if len(main_data) == len(keys):
            return main_data
        main_data.update(data)
        # update the current keys to wait for the remaining keys
        _current_keys = list(set(keys) - set(main_data.keys()))
        time.sleep(frequency)
    raise DatastoreKeyNotFoundError(
        f"Keys {keys} were not found in the datastore after {max_wait_time} seconds."
    )


@contextmanager
def task_sync_barrier(
    barrier_name,
    datastore: DeepspeedDatastore,
    keys: List[str],
    max_wait_time=600,
    frequency=0.1,
    description=None,
):
    """
    A context manager that waits for keys to be written to the datastore and acts like a distributed-barrier.
    When multiple tasks are running in parallel, this context manager can be used to ensure that all tasks
    can wait on a certain keys to be written to the datastore. If the keys are not written to the datastore
    after `max_wait_time` seconds, a `BarrierTimeoutException` error is raised. This way only once all the keys
    are written to the datastore, the tasks will proceed. This context manager is used to ensure that all tasks
    are in sync before proceeding to the next step.

    Args:
        barrier_name (str): The name of the barrier. Used for debugging purposes.
        datastore (DeepspeedDatastore)
        keys (List[str]): The keys to wait for in the datastore.
        max_wait_time (float): The maximum time to wait for the keys to be written to the datastore.
        frequency (float): The frequency to check the datastore for the keys.
        description (str): A description of the barrier. Used for debugging purposes.
    """
    try:
        data = wait_for_key_data(datastore, keys, max_wait_time, frequency)
    except DatastoreKeyNotFoundError:
        raise BarrierTimeoutException(barrier_name, description)
    else:
        yield data



from metaflow.unbounded_foreach import UBF_CONTROL
import subprocess
import socket
import time
import os
from .exceptions import (
    SSHKeyGenException,
    SSHServiceRestartException,
    SSHScanUnsuccessfulException,
)
from metaflow import current

from .datastore import DeepspeedDatastore, task_sync_barrier
from .constants import (
    HOSTFILE_IP_KEY,
    HOSTFILE,
    MPI_PUBLIC_KEY_PATH,
)


class KeyPaths:
    Control = lambda: f"{MPI_PUBLIC_KEY_PATH}/control"
    Worker = lambda x: f"{MPI_PUBLIC_KEY_PATH}/worker/{x}"
    Host = lambda x: f"{HOSTFILE_IP_KEY}/{x}"


def create_and_push_ssh_keys(datastore: DeepspeedDatastore, ubf_context, node_index):
    """
    Create SSH keys and push the public key to the datastore.
    Also write the IP of the host to the datastore.
    """
    if ubf_context == UBF_CONTROL:
        key_push_path = KeyPaths.Control()
    else:
        key_push_path = KeyPaths.Worker(node_index)

    ssh_dir = os.path.expanduser("~/.ssh")
    if not os.path.exists(ssh_dir):
        os.makedirs(ssh_dir)

    os.chmod(ssh_dir, 0o700)  # TODO : Fix me : rest permissioning correct.
    # TODO : handle case where id_rsa is present
    # TODO: handle case where

    _key_name = "id_rsa"
    # Generate host key if it doesn't exist! NEVER overwrite existing ssh keys of some machine
    if not os.path.exists(os.path.join(ssh_dir, _key_name)):
        try:
            result = subprocess.run(
                [
                    "ssh-keygen",
                    "-t",
                    "rsa",
                    "-f",
                    os.path.join(ssh_dir, _key_name),
                    "-N",
                    "",
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise SSHKeyGenException(
                cmd=e.cmd,
                exception=e.stderr,
                worker_type=ubf_context,
                node_index=node_index,
            )

    if not os.path.exists(os.path.join(ssh_dir, "id_rsa.pub")):
        raise SSHKeyGenException(
            cmd="ssh-keygen",
            exception="Public key not found",
            worker_type=ubf_context,
            node_index=node_index,
        )

    # Move public key to S3
    with open(os.path.join(ssh_dir, "id_rsa.pub"), "r") as f:
        datastore.put(key_push_path, f.read())

    # Write the IP of the host to the datastore
    host_key = KeyPaths.Host(node_index)
    datastore.put(host_key, get_my_ip())

    return ssh_dir


def _control_task_ssh_setup(
    datastore,
    ssh_dir,
    world_size,
    max_wait_time=600,
    frequency=0.1,
):
    # All worker indexes start from 1
    _worker_keys = [KeyPaths.Worker(i) for i in range(1, world_size)]
    _lock_args = {
        "description": "Waiting for worker tasks to write public keys to datastore",
        "max_wait_time": max_wait_time,
        "frequency": frequency,
    }
    with task_sync_barrier(
        "ControlTaskPublicKeySync", datastore, _worker_keys, **_lock_args
    ) as data:
        # append all public keys to authorized_keys file
        with open(os.path.join(ssh_dir, "authorized_keys"), "a") as g:
            for p in _worker_keys:
                obj = data[p]
                g.write(obj.text)
            # add self to keys too
            with open(os.path.join(ssh_dir, "id_rsa.pub"), "r") as f:
                g.write(f.read())


def _worker_task_ssh_setup(datastore, ssh_dir, max_wait_time=600, frequency=0.1):
    control_key_path = KeyPaths.Control()
    _lock_args = {
        "description": "Waiting for control task to write public key to datastore",
        "max_wait_time": max_wait_time,
        "frequency": frequency,
    }
    with task_sync_barrier(
        "WorkerTaskPublicKeySync", datastore, [control_key_path], **_lock_args
    ) as data:
        with open(os.path.join(ssh_dir, "authorized_keys"), "a") as g:
            g.write(data[control_key_path].text)


def update_ssh_config_and_restart_sshd(ssh_config_options, ubf_context, node_index, use_sudo=False):
    # This is un-used but can be used later if necessary.
    with open("/etc/ssh/sshd_config", "a") as f:
        f.write("\n".join(ssh_config_options))

    # TODO : Here we require sudo access and sshd service restart. This should be done better!
    try:
        result = subprocess.run(
            (["sudo"] if use_sudo else []) + ["service", "ssh", "restart"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as e:
        raise SSHServiceRestartException(
            cmd=e.cmd,
            exception=e.stderr,
            worker_type=ubf_context,
            node_index=node_index,
        )

    return True


def get_my_ip():
    return socket.gethostbyname(socket.gethostname())


def host_file_sync(datastore, world_size, max_wait_time=600, frequency=0.1):
    host_key_paths = [KeyPaths.Host(i) for i in range(world_size)]
    description = "Waiting for all workers to write their IPs to the datastore"
    with task_sync_barrier(
        "HostFileSync",
        datastore,
        host_key_paths,
        max_wait_time=max_wait_time,
        frequency=frequency,
        description=description,
    ) as data:
        hosts = []
        for p in host_key_paths:
            hosts.append(data[p].text)
        return hosts


def _write_hostfile(hosts, n_slots):
    my_ip = get_my_ip()
    with open(HOSTFILE, "a") as f:
        for h in hosts:
            if h == my_ip:
                h = "127.0.0.1"
            # slots is MPI lingo which is used with deepspeed when we do passwordless ssh connections between jobs
            # slots correlates to the number of processes of user code we wish to run on that instance.
            f.write("%s slots=%s\n" % (h, n_slots))


def _scan_cmd(host):
    return ["ssh-keyscan", "-H", host]


def scan_one_host(host, max_retries=5, retry_delay=5):
    curr_retry = 0
    error_log = None
    while curr_retry < max_retries:
        try:
            result = subprocess.run(
                _scan_cmd(host),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            # TODO [LOGGING] : make the error visible to the user.
            curr_retry += 1
            if e.stderr:
                error_log = e.stderr.decode("utf-8")
            time.sleep(retry_delay)
            continue
        else:
            return result.stdout.decode("utf-8"), None
    return None, error_log  # Only when it failed all tries.


def scan_all_hosts(hosts, ubf_context, node_index, max_retries=5, retry_delay=5):
    host_keys = []
    for host in hosts:
        output, error_log = scan_one_host(host, max_retries, retry_delay)
        if output is None:
            raise SSHScanUnsuccessfulException(error_log, ubf_context, node_index, host)
        host_keys.append(output)
    return host_keys


def setup_known_hosts(hosts, ubf_context, node_index):
    """
    create the known_hosts file with the public keys of all the hosts.
    """
    my_ip = get_my_ip()
    _hosts = [h for h in hosts if h != my_ip]
    _hosts += ["127.0.0.1"]
    host_keys = scan_all_hosts(
        _hosts, ubf_context, node_index, max_retries=5, retry_delay=5
    )
    known_hosts_path = os.path.expanduser("~/.ssh/known_hosts")
    with open(known_hosts_path, "a") as f:
        for key in host_keys:
            f.write(key)


def setup_mpi_env(
    flow_datastore,
    ubf_context: str = UBF_CONTROL,
    all_nodes_started_timeout: int = 600,
    n_slots: int = 1,
    polling_frequency: float = 0.1,
    use_sudo: bool = False,
):
    """
    1. create ssh keys on each host
    2. based on the type of task do the following :
        - For control task wait for the workers to write the keys in the datastore .
            - Once all keys are written, set them on the authorized hosts file.
        - For the Worker tasks, wait for the control task key to be available.
            - Once it is the write the public key into authorized hosts file.
    """
    # NOTE: Where jobset for @kuberentes + @parallel can automate the sshd port opening,
    # AWS Batch will require security groups applied to the compute env for this.

    datastore = DeepspeedDatastore(
        flow_datastore=flow_datastore, pathspec=current.pathspec
    )
    node_index, world_size = current.parallel.node_index, current.parallel.num_nodes

    ssh_dir = create_and_push_ssh_keys(datastore, ubf_context, node_index)

    if ubf_context == UBF_CONTROL:
        _control_task_ssh_setup(
            datastore,
            ssh_dir,
            world_size,
            max_wait_time=all_nodes_started_timeout,
            frequency=polling_frequency,
        )
    else:
        _worker_task_ssh_setup(
            datastore,
            ssh_dir,
            max_wait_time=all_nodes_started_timeout,
            frequency=polling_frequency,
        )

    # At this point nearly all workers would be running the same line of code since
    # they are all aware of the other worker's public keys.
    # Todo : See if we can find a way to get rid of permissioning setting here.
    os.chmod(os.path.join(ssh_dir, "authorized_keys"), 0o600)

    # Instead of doing sshd restart, instead update the ssh_config file.
    # The ssh config can hold `PubKeyAuthentication` and `RSAAuthentication` set for
    # every IP address created from all the workers.

    # Core question : is the sshd service restart done because there is a change in the id_rsa/authorizedhosts
    # along with ssh_d or because there is only change in sshd_config file. If it is just the sshd change then we can get away without restarting the service
    # by just changing the ssh_config file with the right hosts.

    # enable passwordless ssh and done with SUDO!
    # This is not the best way but we can contribute more time
    # if the usage of this is more frequent and requires rootless environments.
    update_ssh_config_and_restart_sshd(
        [
            "PubKeyAuthentication yes",
            "RSAAuthentication yes",
        ],
        ubf_context,
        node_index,
        use_sudo=use_sudo
    )
    # Write all the IP's shared by the workers to the hostfile.

    hosts = host_file_sync(
        datastore,
        world_size,
        max_wait_time=all_nodes_started_timeout,
        frequency=polling_frequency,
    )
    _write_hostfile(hosts, n_slots)

    # At this point all hosts should be accessible so accordingly
    # setup the known hosts file.
    setup_known_hosts(hosts, ubf_context, node_index)

    return hosts



from metaflow.exception import MetaflowException
from collections import defaultdict


class CardDecoratorInjector:
    """
    Mixin Useful for injecting @card decorators from other first class Metaflow decorators.
    """

    _first_time_init = defaultdict(dict)

    @classmethod
    def _get_first_time_init_cached_value(cls, step_name, card_id):
        return cls._first_time_init.get(step_name, {}).get(card_id, None)

    @classmethod
    def _set_first_time_init_cached_value(cls, step_name, card_id, value):
        cls._first_time_init[step_name][card_id] = value

    def _card_deco_already_attached(self, step, card_id):
        for decorator in step.decorators:
            if decorator.name == "card":
                if (
                    decorator.attributes["id"]
                    and card_id in decorator.attributes["id"]
                ):
                    return True
        return False

    def _get_step(self, flow, step_name):
        for step in flow:
            if step.name == step_name:
                return step
        return None

    def _first_time_init_check(self, step_dag_node, card_id):
        """ """
        return not self._card_deco_already_attached(step_dag_node, card_id)

    def attach_card_decorator(
        self,
        flow,
        step_name,
        card_id,
        card_type,
        refresh_interval=5,
    ):
        """
        This method is called `step_init` in your StepDecorator code since 
        this class is used as a Mixin
        """
        from metaflow import decorators as _decorators
        
        if not all([card_id, card_type]):
            raise MetaflowException(
                "`INJECTED_CARD_ID` and `INJECTED_CARD_TYPE` must be set in the `CardDecoratorInjector` Mixin"
            )

        step_dag_node = self._get_step(flow, step_name)
        if (
            self._get_first_time_init_cached_value(step_name, card_id) is None
        ):  # First check class level setting.
            if self._first_time_init_check(step_dag_node, card_id):
                self._set_first_time_init_cached_value(step_name, card_id, True)
                _decorators._attach_decorators_to_step(
                    step_dag_node, ["card:type=%s,id=%s,refresh_interval=%s" % (card_type, card_id, str(refresh_interval))]
                )
            else:
                self._set_first_time_init_cached_value(step_name, card_id, False)



from metaflow.exception import MetaflowException

# from metaflow import Run as MetaflowRun
from typing import List, Dict, Union
import subprocess
import socket
import json
import sys
import os
from .exceptions import DeepspeedException
from .datastore import DeepspeedDatastore
from .status_notifier import (
    TaskStatusNotifier,
    HeartbeatThread,
    wait_for_task_completion,
    HeartbeatTimeoutException,
    TaskFailedException,
)
from .constants import (
    HOSTFILE,
    DEEPSPEED_ENV_FILE
)


def _get_path(local_output_dir, cloud_output_dir, node_index):
    filepath_tuples = []
    for path, subdirs, files in os.walk(local_output_dir):
        for fname in files:
            filepath_tuples.append(
                (
                    f"{cloud_output_dir}/{str(node_index)}/{os.path.relpath(os.path.join(path, fname), local_output_dir)}",
                    os.path.join(path, fname),
                )
            )
    return filepath_tuples


def _dict_to_args(d: Dict[str, str]) -> List[str]:
    data = []
    for k, v in d.items():
        if v == "" or v is None:
            data.append(f"--{k}")
        else:
            data.extend([f"--{k}", json.dumps(v).replace('"', "")])
    return data


class DeepspeedExecutor:
    """
    Instances of the DeepspeedExecutor class are responsible for launching the Deepspeed command. There is one per Metaflow @step annotated with @deepspeed.
    DeepspeedExecutor consumes a list of host IP addresses corresponding to batch job nodes or kubernetes pods and information about how many processes to run on each.
    It assumes the passwordless ssh communication is setup on the all nodes without any user intervention.

    The Deepspeed decorator, which users specify in a Metaflow num_parallel task with @deepspeed, attaches an instance of this class to current.deepspeed.
        Using current.deepspeed.run, users can run the same Deepspeed launch command they would independently of Metaflow.
        This class will handle opening the subprocess, and ensuring other typical Metaflow functionality works as expected.
    """

    def __init__(
        self,
        hosts: List[str],
        n_slots_per_host: int = 1,
        is_gpu: bool = False,
        worker_polling_freq: int = 10,
        flow_datastore = None,
    ) -> None:
        self.is_gpu = is_gpu
        self.n_slots_per_host = n_slots_per_host
        self.hosts = hosts
        self.worker_polling_freq = worker_polling_freq
        self._flow_datastore = flow_datastore
        self._heartbeat_thread = None  # This is only set on the control task

    def _exec_cmd(
        self,
        deepspeed_args: Union[List[str], Dict[str, str]] = [],
        entrypoint: str = None,
        entrypoint_args: Union[List[str], Dict[str, str]] = [],
    ):
    
        # Container to build up the command to be run in a subprocess.
        cmd = ["deepspeed"]
        if type(deepspeed_args) == dict:
            deepspeed_args = _dict_to_args(deepspeed_args)

        # Construct the deepspeed distributed args.
        if "--hostfile" not in deepspeed_args:
            deepspeed_args.extend(["--hostfile", HOSTFILE])
        if "--num_nodes" not in deepspeed_args:
            deepspeed_args.extend(["--num_nodes", str(len(self.hosts))])
        if self.is_gpu and "--num_gpus" not in deepspeed_args:
            deepspeed_args.extend(["--num_gpus", str(self.n_slots_per_host)])
        if "--master_addr" in deepspeed_args:
            raise MetaflowException(
                "Do not specify the --master_addr in your current.run.deepspeed args. Metaflow will set this for you."
            )
        my_ip = socket.gethostbyname(socket.gethostname())
        deepspeed_args.extend(["--master_addr", my_ip])
        cmd.extend(deepspeed_args)

        # Construct rest of command starting with the entrypoint.
        if entrypoint is not None:
            cmd.append(entrypoint)
        else:
            raise MetaflowException(
                "current.deepspeed.run(..., entrypoint=<SCRIPT>, ...) arg must be specified."
            )
        if entrypoint_args is not None and isinstance(entrypoint_args, dict):
            cmd.extend(_dict_to_args(entrypoint_args))
        elif entrypoint_args is not None and isinstance(entrypoint_args, list):
            cmd.extend(entrypoint_args)

        # Write env_var=value to file.
        # Deepspeed automatically looks for this file to prepend the variables to its launcher command.
        env = os.environ.copy()
        if len(self.hosts) > 1:
            with open(DEEPSPEED_ENV_FILE, "w") as f:
                for k, v in os.environ.items():
                    if (
                        k == "METAFLOW_INIT_SCRIPT"
                    ):  # TODO: Remove the MF_PARALLEL variables from here since it will be passed down to workers via SSH.
                        continue  # Don't pass these to deepspeed. Because of how deepspeed reads env vars and sets them in runner commands.
                    else:
                        json_string = json.dumps(v)
                        f.write(f"{k}={json_string}\n")
        
        elif len(self.hosts) == 1:  # if multi-node, deepspeed does this itself.
            curr_path = os.path.abspath(".")
            env["PYTHONPATH"] = curr_path

        # Launch the Deepspeed run.
        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,  # pipe to Metaflow stdout. TODO: How to handle progress bar buffering like TQDM.
            stderr=subprocess.STDOUT,
            env=env
        ) as process:
            while process.poll() is None:
                stdout = process.stdout.read1()
                try:
                    text = stdout.decode("utf-8")
                except UnicodeDecodeError:
                    text = ""
                print(
                    text, end="", flush=True
                )  # TODO: other mechanism to flush buffer, especially in progress bar case.

            if process.returncode != 0:
                if process.stderr is not None:
                    return False, process.stderr.read().decode("utf-8")
                else:
                    return False, None
            return True, None

    def _resolve_storage_paths(
        self, push_results_dir_to_cloud, datastore, local_output_dir, cloud_output_dir
    ) -> str:
        if push_results_dir_to_cloud and (
            datastore._backend.TYPE != "s3" and datastore._backend.TYPE != "azure"
        ):
            raise MetaflowException(
                "current.deepspeed.run must use S3 or Azure Blob Storage as a datastore if push_results_dir_to_cloud is True. You are using %s."
                % datastore._backend.TYPE
            )
        elif push_results_dir_to_cloud:
            # TODO: Annoying place for this check. Consider moving the S3 push args into the decorator itself, so can be checked at flow init instead.
            if local_output_dir is None:
                raise MetaflowException(
                    "current.deepspeed.run must specify local_output_dir if push_results_dir_to_s3 is True"
                )
            elif cloud_output_dir is None:
                if local_output_dir.startswith("/"):
                    cloud_output_dir = local_output_dir[1:]
                else:
                    cloud_output_dir = local_output_dir
        return cloud_output_dir

    def _control_node_task(
        self, datastore: DeepspeedDatastore, deepspeed_args, entrypoint, entrypoint_args
    ):
        """
        The control task will run the `deepspeed` command using all the arguments provided
        to the `DeepSpeedExecutor.run` method. The control task will also publish heartbeats
        to the datastore every 3 seconds to indicate that it is still running. These heartbeats
        are used by the worker tasks to monitor the control task's status and essentially keep the workers alive.

        Since the `DeepSpeedExecutor` is an abstraction over the `deepspeed` command, we will capture the stderr
        when we call the `deepspeed` command and pass it down to the `DeepspeedException` if the `deepspeed` command fails.
        """
        # set the status that the control node is Up
        _status_notifier = TaskStatusNotifier(datastore)
        _status_notifier.running(0)
        # start the heartbeat thread that writes the status to the datastore every 3 seconds
        self._heartbeat_thread = HeartbeatThread(
            _status_notifier, node_index=0, heartbeat_interval=3
        )
        self._heartbeat_thread.start()
        # run the deepspeed command
        # TODO : Fix the subprocess reading and logging logic.
        success_status, stderr = self._exec_cmd(
            deepspeed_args, entrypoint, entrypoint_args
        )
        # stop the heartbeat thread upon failure / success of control task.
        self._heartbeat_thread.stop()
        # sets the status of the control node to finished or failed
        if success_status:
            _status_notifier.finished(0)
        else:
            _status_notifier.failed(0)
            msg = f"The `deepspeed` command running on the control task has crashed. \n\n[stderr]: {str(stderr)}"
            raise DeepspeedException(msg)

    def _worker_node_task(
        self, datastore: DeepspeedDatastore, node_index: int, heartbeat_timeout=60 * 10
    ):
        """
        The worker tasks will poll for the control task's heartbeat and do nothing else.
        Given that the tasks are run using MPI/ssh, the control task ends up controlling the worker directly (via ssh)
        and metaflow doesn't need to do anything else.

        Any failure in the worker's entry-point script will result in the failure at the control task level.
        The only way a worker will end up failing will be :
            - The control task fails (Which can happen the worker's entry-point script fails, resulting in the control task failing.)
            - if the duration of the last heartbeat crosses the `heartbeat_timeout`

        Hence the control flow on the worker only monitors the heartbeat/statues set from the control task.
        """
        # TODO : Make heartbeat timeout configurable
        _status_notifier = TaskStatusNotifier(datastore)
        # Worker task statuses are only for bookkeeping.
        # They are not used by the control task in any way.
        _status_notifier.running(node_index)
        try:
            # Poll the control task's heartbeat and fail if control task fails
            # or if the heartbeat interval crosses the threshold.
            wait_for_task_completion(
                _status_notifier, node_index=0, heartbeat_timeout=heartbeat_timeout
            )
            _status_notifier.finished(node_index)
        except HeartbeatTimeoutException:
            _status_notifier.failed(node_index)
            raise DeepspeedException(
                f"Control task heartbeat timed out. Control task has not published a heartbeat for {heartbeat_timeout} seconds."
            )
        except TaskFailedException:
            _status_notifier.failed(node_index)
            raise DeepspeedException("Control task reported failure.")

    def run(
        self,
        deepspeed_args: Union[List[str], Dict[str, str]] = [],
        entrypoint: str = None,
        entrypoint_args: Union[List[str], Dict[str, str]] = [],
        push_results_dir_to_cloud: bool = False,
        local_output_dir: str = None,
        cloud_output_dir: str = None,  # TODO: Remove the stuff about `push_results_dir_to_cloud`
    ) -> Union[str, None]:
        """upon completion returns the final path of the `cloud_output_dir` if `push_results_dir_to_cloud` is set to true"""
        from metaflow import current

        node_index = current.parallel.node_index or 0
        datastore = DeepspeedDatastore(
            flow_datastore=self._flow_datastore, pathspec=current.pathspec
        )

        # Resolve storage paths
        cloud_output_dir = self._resolve_storage_paths(
            push_results_dir_to_cloud, datastore, local_output_dir, cloud_output_dir
        )

        # Run the distributed job
        if node_index == 0:  # control node
            self._control_node_task(
                datastore, deepspeed_args, entrypoint, entrypoint_args
            )
        else:  # worker node
            self._worker_node_task(datastore, node_index)

        # Push results to S3
        if push_results_dir_to_cloud:
            if not os.path.exists(local_output_dir):
                print(
                    f"Deepspeed process completed, and local_output_dir `{local_output_dir}` does not exist, skipping push to datastore.",
                    file=sys.stderr,
                )
                return None
            paths = datastore.put_files(
                _get_path(local_output_dir, cloud_output_dir, node_index)
            )
            print(
                f"Pushed {len(paths)} files to {datastore._backend.TYPE} at {cloud_output_dir}",
                file=sys.stderr,
            )
            return datastore.get_datastore_file_location(cloud_output_dir)

    @classmethod
    def for_single_node(cls, flow, use_gpu=False, n_slots=1):
        from .mpi_setup import setup_mpi_env
        flow_datastore = flow._datastore.parent_datastore
        hosts = setup_mpi_env(flow_datastore=flow_datastore, n_slots=n_slots, )
        executor = cls(
            hosts=hosts, 
            n_slots_per_host=n_slots,
            is_gpu=use_gpu,
            flow_datastore=flow_datastore
        )
        return executor


import os
from metaflow.cards import (
    Markdown,
    Table,
    ProgressBar,
)
from metaflow.decorators import StepDecorator
import math
from metaflow.metaflow_current import current
from ..card_utilities.async_cards import CardRefresher, AsyncPeriodicRefresher
from ..card_utilities.extra_components import ArtifactTable, LineChart
from ..card_utilities.injector import CardDecoratorInjector
from .constants import DEFAULT_FILE_NAME, DEFAULT_PROFILER_FILE_NAME


def json_to_artifact_table(data):
    return ArtifactTable(data)


class HuggingfaceModelMetricsRefresher(CardRefresher):
    CARD_ID = "training_metrics"

    def __init__(self) -> None:
        self._metrics_charts = {}
        self._runtime_info_table = {}
        self._progress_bars = {}
        self._last_updated_on = None
        self._already_rendered = False

    def render_card_fresh(self, current_card, data):
        self._already_rendered = True
        current_card.clear()
        current_card.append(
            Markdown(
                "## Huggingface Model Training Metrics \n## %s [Attempt:%s]"
                % (current.pathspec, current.retry_count),
            )
        )
        self._last_updated_on = Markdown(
            f"_Last data-update on: {data['created_on']}_",
        )
        current_card.append(self._last_updated_on)
        current_card.append(
            Markdown(
                "## Training Progress",
            )
        )
        self._progress_bars["epoch"] = ProgressBar(
            label="Epoch",
            max=data["training_state"]["num_train_epochs"],
            value=round(data["training_state"]["epoch"], 2),
        )
        self._progress_bars["global_step"] = ProgressBar(
            label="Global Step",
            max=data["training_state"]["max_steps"],
            value=data["training_state"]["global_step"],
        )
        _steps_per_epoch = int(
            data["training_state"]["max_steps"]
            / data["training_state"]["num_train_epochs"]
        )
        _steps_in_epoch_current = int(
            data["training_state"]["global_step"]
            - (_steps_per_epoch * math.floor(data["training_state"]["epoch"]))
        )
        self._progress_bars["epoch_steps"] = ProgressBar(
            label="Steps in Epoch",
            max=_steps_per_epoch,
            value=_steps_in_epoch_current,
        )
        current_card.append(
            Table(
                data=[
                    [
                        self._progress_bars["epoch"],
                    ],
                    [
                        self._progress_bars["global_step"],
                    ],
                    [
                        self._progress_bars["epoch_steps"],
                    ],
                ]
            )
        )
        current_card.append(
            Markdown(
                "## Metrics",
            )
        )
        for metric in data["metrics"]:
            if metric not in self._metrics_charts:
                self._metrics_charts[metric] = LineChart(
                    title=metric,
                    xtitle="Step",
                    ytitle=metric,
                    x_name="step",
                    y_name="value",
                    with_params=False,
                    x_axis_temporal=False,
                )
            self._metrics_charts[metric].spec["data"]["values"] = data["metrics"][
                metric
            ]
            current_card.append(self._metrics_charts[metric], id=metric)
        if len(data["runtime_info"]) > 0:
            current_card.append(
                Markdown(
                    "## Runtime Information",
                )
            )
            for k in data["runtime_info"]:
                self._runtime_info_table[k] = Markdown(str(data["runtime_info"][k]))
            current_card.append(
                Table(
                    data=[
                        [Markdown("**%s**" % k), self._runtime_info_table[k]]
                        for k in self._runtime_info_table
                    ]
                )
            )

        current_card.refresh()

    def on_startup(self, current_card):
        current_card.append(
            Markdown(
                "# Huggingface Model Training Metrics\n## %s [Attempt:%s]"
                % (current.pathspec, current.retry_count),
            )
        )
        current_card.append(
            Markdown(
                "_waiting for data to appear_",
            )
        )
        current_card.refresh()

    def on_error(self, current_card, error_message):
        if isinstance(error_message, FileNotFoundError):
            return

        if not self._already_rendered:
            current_card.clear()
            current_card.append(
                Markdown(
                    f"## Error: {str(error_message)}",
                )
            )
            current_card.refresh()

    def update_only_components(self, current_card, data_object):
        self._last_updated_on.update(
            f"_Last data-update on: {data_object['created_on']}_"
        )
        if len(data_object["runtime_info"]) > 0:
            for k in data_object["runtime_info"]:
                self._runtime_info_table[k].update(str(data_object["runtime_info"][k]))

        for metric in data_object["metrics"]:
            self._metrics_charts[metric].spec["data"]["values"] = data_object[
                "metrics"
            ][metric]

        self._progress_bars["epoch"].update(
            round(data_object["training_state"]["epoch"], 2),
        )
        self._progress_bars["global_step"].update(
            data_object["training_state"]["global_step"],
        )
        _steps_per_epoch = int(
            data_object["training_state"]["max_steps"]
            / data_object["training_state"]["num_train_epochs"]
        )
        _steps_in_epoch_current = int(
            data_object["training_state"]["global_step"]
            - (_steps_per_epoch * math.floor(data_object["training_state"]["epoch"]))
        )
        self._progress_bars["epoch_steps"].update(_steps_in_epoch_current)
        current_card.refresh()

    def on_update(self, current_card, data_object):
        data_object_keys = set(data_object.keys())
        if len(data_object_keys) == 0:
            return
        if len(self._metrics_charts) == 0:
            self.render_card_fresh(current_card, data_object)
            return
        elif len(data_object["metrics"]) != len(self._metrics_charts):
            self.render_card_fresh(current_card, data_object)
            return
        elif (
            len(data_object["runtime_info"]) > 0 and len(self._runtime_info_table) == 0
        ):
            self.render_card_fresh(current_card, data_object)
            return
        else:
            self.update_only_components(current_card, data_object)
            return


class HuggingfaceModelCardRefresher(CardRefresher):
    CARD_ID = "training_variables"

    def __init__(self) -> None:
        self._rendered = False

    def render_card_fresh(self, current_card, data):
        self._rendered = True
        current_card.clear()
        current_card.append(
            Markdown(
                "# Huggingface Model Training Configuration \n## %s [Attempt:%s]"
                % (current.pathspec, current.retry_count),
            )
        )
        current_card.append(
            Markdown(
                "## Trainer Configuration",
            )
        )
        current_card.append(json_to_artifact_table(data["trainer_configuration"]))
        current_card.append(
            Markdown(
                "## Model Configuration",
            )
        )
        current_card.append(json_to_artifact_table(data["model_config"]))
        current_card.refresh()

    def on_startup(self, current_card):
        current_card.append(
            Markdown(
                "# Huggingface Model Training Config\n## %s[Attempt:%s]"
                % (current.pathspec, current.retry_count),
            )
        )
        current_card.append(
            Markdown(
                "_waiting for data to appear_",
            )
        )
        current_card.refresh()

    def on_error(self, current_card, error_message):
        if isinstance(error_message, FileNotFoundError):
            return
        if not self._rendered:
            current_card.clear()
            current_card.append(
                Markdown(
                    f"## Error: {str(error_message)}",
                )
            )
            current_card.refresh()

    def on_update(self, current_card, data_object):
        if self._rendered:
            return
        data_object_keys = set(data_object.keys())
        if len(data_object_keys) == 0:
            return
        required_keys = set(
            [
                "trainer_configuration",
                "model_config",
            ]
        )
        if not data_object_keys.issuperset(required_keys):
            return

        self.render_card_fresh(current_card, data_object)


class HuggingFaceCardDecorator(StepDecorator, CardDecoratorInjector):

    name = "huggingface_card"

    def step_init(self, flow, graph, step_name, decorators, environment, flow_datastore, logger):
        self.attach_card_decorator(
            flow,
            step_name,
            HuggingfaceModelCardRefresher.CARD_ID,
            "blank",
            refresh_interval=10,
        )
        self.attach_card_decorator(
            flow,
            step_name,
            HuggingfaceModelMetricsRefresher.CARD_ID,
            "blank",
            refresh_interval=0.5
        )
    
    def task_decorate(self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context):
        def _wrapped_step_func(*args, **kwargs):
            async_refresher_model_card = AsyncPeriodicRefresher(
                HuggingfaceModelCardRefresher(),
                updater_interval=3,
                collector_interval=2,
                file_name=DEFAULT_FILE_NAME,
            )
            async_refresher_metrics = AsyncPeriodicRefresher(
                HuggingfaceModelMetricsRefresher(),
                updater_interval=1,
                collector_interval=0.5,
                file_name=DEFAULT_FILE_NAME,
            )
            try:
                async_refresher_model_card.start()
                async_refresher_metrics.start()
                return step_func(*args, **kwargs)
            finally:
                async_refresher_model_card.stop()
                async_refresher_metrics.stop()

        return _wrapped_step_func
        


import time
import os

import metaflow
from metaflow import current
from metaflow.unbounded_foreach import UBF_CONTROL
from metaflow.plugins.parallel_decorator import (
    ParallelDecorator,
    _local_multinode_control_task_step_func,
)

from .executor import DeepspeedExecutor
from .mpi_setup import setup_mpi_env
import os

"""
Control Flow : 
1. Task pre-step sets the UBF Context. 
2. The @parallel decorator mutates the step function based because it task decorates. 
    - This mutation is essential because, we are running the actual user code differently for each task based on the type of task (worker / control).
    - The `@parallel` decorator exposes a `setup_distributed_env` method that is called by the `@deepspeed` decorator to set up the distributed environment.
3. Once `setup_distributed_env` is called, we sets up the SSH connections between the nodes for MPI. 
4. Use code can now call `current.deepspeed.run` to run the training script with deepspeed. 
    - `current.deepspeed.run` only executes the entry-point script on the control task
    - The worker tasks are controlled from the control task using MPI with SSH. 
"""


class DeepspeedDecorator(ParallelDecorator):
    name = "deepspeed"
    # TODO : Safe rename `all_nodes_started_timeout` to something more intuitive.
    defaults = {"all_nodes_started_timeout": 600, "worker_polling_freq": 0.5, "use_sudo": False}
    IS_PARALLEL = True

    requires_passwordless_ssh = (
        True  # modifies how @kubernetes runs command in docker container
    )

    def _setup_current(self, hosts, n_slots_per_host, is_gpu, flow):
        current._update_env(
            {
                "deepspeed": DeepspeedExecutor(
                    hosts=hosts,
                    n_slots_per_host=n_slots_per_host,
                    is_gpu=is_gpu,
                    worker_polling_freq=self.attributes["worker_polling_freq"],
                    flow_datastore=self.flow_datastore,
                )
            }
        )

    def step_init(self, flow, graph, step, decos, environment, flow_datastore, logger):
        super().step_init(flow, graph, step, decos, environment, flow_datastore, logger)
        self.environment = environment
        self.flow_datastore = flow_datastore

        for deco in decos:
            if deco.name in ["resources", "kubernetes", "batch"]:
                if deco.attributes["gpu"]:
                    self.is_gpu = True
                    self.n_slots = deco.attributes["gpu"]
                else:
                    self.is_gpu = False
                    self.n_slots = deco.attributes["cpu"]
            # Since Netflix/metaflow#1775 and Netflix/metaflow#1776 add the
            # port and the `jobsets` implementation without services, we
            # can just go about setting the port to 22 if it is not set.
            # If the older implementation with `requires_passwordless_ssh` is
            # if being used then as a fallback the attribute will still exist.
            if deco.name == "kubernetes":
                if "port" in deco.defaults:
                    deco.attributes["port"] = 22

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self._ubf_context = ubf_context
        super().task_pre_step(
            step_name,
            task_datastore,
            metadata,
            run_id,
            task_id,
            flow,
            graph,
            retry_count,
            max_user_code_retries,
            ubf_context,
            inputs,
        )

    def setup_distributed_env(
        self,
        flow,
    ):
        "Return a list of strings of hostnames of nodes to use for MPI"
        hosts = setup_mpi_env(
            ubf_context=self._ubf_context,
            all_nodes_started_timeout=self.attributes["all_nodes_started_timeout"],
            n_slots=self.n_slots,
            flow_datastore=self.flow_datastore,
            use_sudo = self.attributes.get("use_sudo", False),
        )
        self._setup_current(
            hosts=hosts, n_slots_per_host=self.n_slots, is_gpu=self.is_gpu, flow=flow
        )
        return hosts



from setuptools import setup, find_namespace_packages

version = "0.0.9"

setup(
    name="metaflow-deepspeed",
    version=version,
    description="An EXPERIMENTAL Deepspeed decorator for Metaflow",
    author="Eddie Mattia",
    author_email="eddie@outerbounds.com",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["metaflow_extensions.*"]),
    py_modules=[
        "metaflow_extensions",
    ],
    install_requires=[],
)